/* ----------------------------------------------------------------------------------------
|
| 	[ATMega88 mit SI570, LCD und Drehimpulsgeber]
| 	Copyright (C) November [2011]  [Clemens Verstappen DL3ETW] 
| 
| 
| 	Basierendes Werk "TR2200GX with Silabs Si570 LO control firmware" von 
| 	Wulf-Gerd Traving, DL1FAC 3/2009, unter Verwendung der lcd.c und twimaster.c 
|	Module von Peter Fleury.
|	
| 	Verwendendete Version:
|	AVR Studio		4.16 Build 628  
| 	AVR GNU Compiler Collection (GCC) 4.3.3 	
|
|	Funktionsumfang:
|	
|	Die im Flashmemory abgelegten Frequenzen (32Bit Integer) werden via I2C 
|	an den SI570 ausgegeben und ber das LCD Display mit Frequenz und Kanal 
|	angezeigt. Die Frequenzauswahl erfolgt ber den Drehimpulsgeber. 
|
|	1) Fusebits:
|	Ext. Crystal Osc. 8 MHz; Start-up time PWRDWN/RESET: 16K CK/14 CK + 65 ms
|	bei Verwendung eines 20MHz Quarzes.SPIEN. 
|
|	2) Source Files mit einbinden:
|	lcd.c
|	twimaster.c
|
|	3) Unter Project/Configurations Options/
|	hier "atmega88" anwhlen
|
|-----------------------------------------------------------------------------------------
| License:
| This program is free software; you can redistribute it and/or modify it under
| the terms of the GNU General Public License as published by the Free Software
| Foundation; either version 2 of the License, or (at your option) any later
| version.
| This program is distributed in the hope that it will be useful, but
|
| WITHOUT ANY WARRANTY;
|
| without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
| PURPOSE. See the GNU General Public License for more details.
|
| You should have received a copy of the GNU General Public License along with
| this program; if not, write to the Free Software Foundation, Inc., 51
| Franklin St, Fifth Floor, Boston, MA 02110, USA
|
| http://www.gnu.de/gpl-ger.html
`-----------------------------------------------------------------------------------------*/

#define F_CPU 20000000UL		//  fr Delay

#include "i2cmaster.h"			//	I2C Ansteuerung des SI570 (in Hardware)  				
#include "lcd.h"				//	Ansteuerung des zweireihigen LCD Displays
#include <avr/interrupt.h>		//	Fr Timer und den externen Interrupt 
#include <avr/pgmspace.h>		//	Lesen aus dem Flash, Progammspeicher
#include <stdlib.h>				// 	Routinen zur Zeichenausgabe ber das LCD
#include <string.h>				//	Routine zur Ermittlung der Zeichenlnge
#include <util/delay.h>			//	Routine fr Zeitverzgerung
#include <math.h>				// 	Fr den log10
	
#define SI570ADDR 0xaa			//	I2C Adresse des SI570
#define FOUT0 56.3199			//	Hier die Frequenz (MHz) eingeben, die direkt nach Anlegen
								//	der Versorgungsspannung am SI570 erzeugt wird

const float FDCO_MAX = 5670.0; 					// MHz,  Si57x's fdco range -- DO NOT EDIT
const float FDCO_MIN = 4850.0; 					// MHz
const unsigned long int FDCO_MAXL = 5537109; 	// * 1024 Hz
const unsigned long int FDCO_MINL = 4736328; 	// * 1024 Hz
const unsigned char HS_DIV[6] = {11, 9, 7, 6, 5, 4};

char s[16];
uint8_t sreg,initial_n1,initial_hsdiv,kanal,zeiger=2;
uint8_t zeile1[16], zeile2[16],reg[6];

uint32_t initial_rfreq_long,frequenz;

uint32_t PROGMEM channel_tab [] PROGMEM = {
			0, 	145600000, 145600000,			//R0
			0, 	145612500, 145612500,			//R0X
			1, 	145625000, 145625000,			//R1
			1, 	145637500, 145637500,			//R1X
			2, 	145650000, 145650000,         	//R2       
			2, 	145662500, 145662500,			//R2X
			3, 	145675000, 145675000,         	//R3             
			3, 	145687500, 145687500,			//R3X
			4, 	145700000, 145700000,         	//R4             
			4,	145712500, 145712500,			//R4X
			5,	145725000, 145725000,         	//R5             
			5,	145737500, 145737500,			//R5X
			6,	145750000, 145750000,         	//R6              
			6,	145762500, 145762500,			//R6X
			7,	145775000, 145775000,         	//R7               
			7,	145787500, 145787500,			//R7X
			8,	145800000, 145800000,         	//R8                        
			8,	145325000, 145325000,			//R8B		MHz(F)
			9,	145350000, 145350000,			//R9B		MHz(F)                                
			10,	145375000, 145375000,			//R10		MHz(F)
			11,	145400000, 145400000,			//R11		MHz(F)                                
			12,	145425000, 145425000,			//R12		MHz(F)
			18,	145450000, 145450000,			//R18		MHz(OE)                             
			19,	145475000, 145475000,			//R19		MHz(OE)
};

volatile struct BitVariablen{				//	um Status zu merken bei nderungen
    int kanal_update 	:1;
	int	silabs_detected :1;
}status = {1};


SIGNAL(TIMER1_COMPA_vect) {					// Signal(Timer_Compa_vect) Interrupt jede 1/4 Sekunde
	PORTB ^= (1 << PORTB1);					// Lebenslicht 0.5 Hz
} 
	
SIGNAL(INT1_vect) {							// externer  Interrupt, steigende Flanke von Drehencoder B 
		status.kanal_update = 1;			// nderung in dieser Bitvariable merken um LCD+VCO updaten
		if (PINC & (1 << PC3)) 	{			// Wie ist Pegel von Drehencoder A?
			if (zeiger > 0)					// wenn noch grer eins, dann darf noch decrementiert werden
					zeiger--;				// Drehrichtung gegen den Uhrzeigersinn
			else
			zeiger=23;						// und oben wieder anfangen
			}
		else					{
			if (zeiger < 23)				// Der Pegel von Drehencoder A ist negativ
					zeiger++;				// Drehrichtung mit den Uhrzeigersinn
			else
			zeiger=0;						// und unten wieder anfangen
			}
		sreg = SREG;						// Timer1 aus Null setzen, um Drehen an Impulsgeber zu ermglichen  
		cli();								// Disable interrupts
		TCNT1 = 0x0000;						// Clear Timer1 auf, von null wieder anfangen 
		SREG = sreg;						// Restore global interrupt flag
		sei();
}

int I2C_ByteWrite (unsigned char reg, unsigned char data) {
	if (i2c_start (SI570ADDR|I2C_WRITE)) {	// failed to issue start condition, possibly no device found */
		i2c_stop();
		status.silabs_detected = 0;
		return -1;
	}
	else 	{								// issuing start condition ok, device accessible */
		i2c_write(reg);
		i2c_write(data);
		i2c_stop();
	}
	return 0;
}

int I2C_regsWrite (void) {
    unsigned char i;
	if (i2c_start (SI570ADDR|I2C_WRITE)) 	{ // failed to issue start condition, possibly no device found */
		i2c_stop();
		status.silabs_detected = 0;
		return -1;
	}
	else 	{								// issuing start condition ok, device accessible */
		i2c_write (7);						// load the new values into the device at registers 7 to 12
	    for (i=0; i<sizeof(reg); i++) i2c_write (reg[i]);
		i2c_stop();
	}
	return 0;
}

unsigned char I2C_ByteRead (unsigned char reg) {
	unsigned char ret;
	if (i2c_start (SI570ADDR|I2C_WRITE)) {	 // failed to issue start condition, possibly no device found */
		i2c_stop();
		status.silabs_detected = 0;
		return -1;
	}
	else 	{
		i2c_write (reg);
		i2c_rep_start (SI570ADDR+I2C_READ);  // set device address and read mode
		ret = i2c_readNak ();                // read one byte
		i2c_stop ();
		return ret;
	}
}

void ReadStartUpConfiguration(void)	{		// Reads start-up register contents for RFREQ, HS_DIV, and N1 and calculates the internal crystal frequency (xtal)
	unsigned char counter;					// recall the initial state of the device from nvm into ram
	if (I2C_ByteWrite(135, 0x01) < 0) 	{
		status.silabs_detected = 0;
		return;
	}										// read registers 7 to 12, reg[0] is equivalent to register 7 in the device, reg[5] is register 12 in the device
	for(counter=0; counter<6; counter++) 	{
		reg[counter] = I2C_ByteRead(counter+7);
	}
	initial_hsdiv = ((reg[0] & 0xE0) >> 5) + 4; // HS_DIV conversion, get reg 7 bits 5, 6, 7
												// initial_hsdiv's value could be verified here to ensure that it is one
												// of the valid HS_DIV values from the datasheet.
												// initial_n1 conversion
	initial_n1 = (( reg[0] & 0x1F ) << 2 ) + (( reg[1] & 0xC0 ) >> 6 );  // get reg 7 bits 0 to 4
												// add with reg 8 bits 7 and 8
	if(initial_n1 == 0) 	{
		initial_n1 = 1;
	}
	else if ((initial_n1 & 1) != 0) 	{		// add one to an odd number
		initial_n1 = initial_n1 + 1;			// RFREQ conversion using unsigned long
	}
	initial_rfreq_long = ( reg[1] & 0x3F );
	initial_rfreq_long = (initial_rfreq_long << 8) + ( reg[2] );
	initial_rfreq_long = (initial_rfreq_long << 8) + ( reg[3] );
	initial_rfreq_long = (initial_rfreq_long << 8) + ( reg[4] );
	initial_rfreq_long = (initial_rfreq_long << 6) + ( reg[5] >> 2 ); //ignore lowest two bits to fit into long
}

void silabs_setNewFreq (uint32_t newFrequency) {	// Program a new frequency
	unsigned char counter;
	unsigned char n1;
	unsigned char hsdiv;
	unsigned char validCombo;
	unsigned char reg137, reg135;
	unsigned int divider_max;
	unsigned int curr_div;
	unsigned int rn1;
	float currentFrequency;
	float ratio = 0;
	unsigned long final_rfreq_long;

	if (status.silabs_detected) 	{
		currentFrequency = (float) newFrequency / 1000000.0;	// find dividers (get the max and min divider range for the HS_DIV and N1 combo)
		divider_max = FDCO_MAXL / (newFrequency >> 10);
		curr_div = (FDCO_MINL / (newFrequency >> 10)) + 1;
		validCombo = 0;
		n1 = initial_n1;
		hsdiv = initial_hsdiv;
		while (curr_div <= divider_max) {				// check all the HS_DIV values with the next curr_div
			for(counter=0; counter<6; counter++) {		// get the next possible n1 value
				hsdiv = HS_DIV[counter];
				rn1 = curr_div % hsdiv;
				if (rn1 == 0) {							// no remainder, this is a possible n1 value
					n1 = curr_div / hsdiv;
					if ((n1 == 1) || ((n1 & 1) == 0)) goto foundCombo;
				}										// increment curr_div to find the next divider
		}												// since the current one was not valid
		curr_div += 1;
		}
		foundCombo:
				// if(validCombo == 0) at this point then there's an error
				// in the calculatio. Check if the provided fout0 and fout1
				// are not valid frequencies
				// calculate RFREQ organizing the float variables to save precision;
				// RFREQ is kept as an unsigned long
				// only 32 bits are available in the long format
				// RFREQ in the device has 34 bits of precision
				// only 34 of the 38 bits are needed since RFREQ is between 42.0 and 50.0 for xtal of 114.285MHz (nominal)
		ratio = currentFrequency / FOUT0;
		ratio *= ((float) n1 / (float) initial_n1) * ((float) hsdiv / (float) initial_hsdiv);
		final_rfreq_long = ratio * (float) initial_rfreq_long;
		reg[0] = 0;										// set the top 3 bits of reg 13
		reg[0] = (hsdiv-4) << 5;						// convert new N1 to the binary representation
		if (n1 == 1) n1 = 0;
		else if((n1 & 1) == 0) n1 = n1 - 1; 			//if n1 is even, subtract one
		reg[0] = (reg[0] & 0xE0) | (n1 >> 2);			// set reg 7 bits 0 to 4
		reg[1] = (n1 & 3) << 6;							// set reg 8 bits 6 and 7

		reg[1] |= final_rfreq_long >> 30;				// load new RFREQ into register map
		reg[2] = final_rfreq_long >> 22;
		reg[3] = final_rfreq_long >> 14;
		reg[4] = final_rfreq_long >> 6;
		reg[5] = final_rfreq_long << 2;
		
		do {								// Load the new frequency, first wait for previous setup beeing through
			reg135 = I2C_ByteRead(135);
		} while (reg135 & 0x40);			// get the current state of register 137
		reg137 = I2C_ByteRead(137);			// set the Freeze DCO bit in that register
		I2C_ByteWrite(137, reg137 | 0x10); 	// load the new values into the device at registers 7 to 12
		I2C_regsWrite ();					// get the current state of register 137
		reg137 = I2C_ByteRead(137);			// clear the FZ_DCO bit in that register
		I2C_ByteWrite(137, reg137 & 0xEF);	// set the NewFreq bit, bit will clear itself once the device is ready
		I2C_ByteWrite(135, 0x40);
	}
}

int main(void) 	{ 
	DDRB 	= 0b00000010;					// PB1 ist Lebenslicht, sonst Eingnge
	DDRC 	= 0b00110000;					// PortC ist Input, bis auf SCL, SDA	
	
	OCR1A	= 0x07a1;						// Compare Match Register Laden, 4C4B = 1s ;2626 1313 = 250ms, 7A1=100ms 
	TCCR1B	= 0b00001101;					// WGM12 setzen, clear on compare match, divide by 1024
	TIMSK1	= 0b00000010;					// Timer1 compare match 1 interrupt enable

	ADMUX	=0b11000000;					// A/D initialisieren, trigger free run
	ADCSRA	=0b11100100;					//
  	
	EICRA	|= (1<<ISC11)|(1<<ISC10);		// Steigende Flanke fr externen Interrupt 1, Drehimpulsgeber B
	EIMSK	|= (1<<INT1);					// externen Interrupt1 freigeben	

 
	lcd_init(LCD_DISP_ON);					// LCD initialiesieren
 	lcd_clrscr();							// und lschen
	lcd_gotoxy(0,0);						// 	dies in die zweite Zeile des LCD
	lcd_puts("DISPLAY READY   ");	
	_delay_ms (1000.0);						// 	hier mal warten

	i2c_init();                             // initialize I2C library
    status.silabs_detected = 1;
    ReadStartUpConfiguration();				// hier erster Zugriff auf SI570
	lcd_gotoxy(0,0);						
	lcd_puts("I2C READY       ");			//  wenn Antwort, dann ready
	_delay_ms (1000.0);						// 	hier mal warten

	sei ();									// 	nun Interrupts freigeben

	for (;;)  	{

		if (status.kanal_update) 	{		// 	nur abarbeiten, wenn neuer Kanal angewhlt
			unsigned char i;
			kanal = pgm_read_dword (&channel_tab [zeiger*3]);		// den Kanal aus PGM lesen
			utoa(kanal,s,10);				// 	16Bit unsign Integer in Ascii 
			for(i = 0; s[i] && i < 3; i++)	// 	maximal 3 Stellen kopieren in Zeile1
			zeile1[i]=s[i];
			zeile1[i++]=' ';	
   			zeile1[i++]=' ';		
  			zeile1[i++]=' ';		
   			zeile1[i++]=' ';
 			zeile1[i]=0;						// abschlieende Null
			lcd_gotoxy(0,0);				
			lcd_puts("Kanal:          ");		//  alte Werte lschen
			lcd_gotoxy(6,0);	
			for(i = 0; zeile1[i] && i < 8; i++)	// 	auslesen bis NULL, 8 Zeichen nicht mehr
  			lcd_putc(zeile1[i]);
												// Zweite Zeile
			frequenz = pgm_read_dword (&channel_tab [zeiger*3+1]); 	// nun zugehrige Frequenz einlesen
											
			ultoa(frequenz,s,10);				// 	32Bit unsign Integer in Ascii 
			for(i = 0; i < 10; i++)				// 	maximal 10Stellen kopieren in Zeile2
			zeile2[i]=s[i];
			lcd_gotoxy(0,1);					// 	dies in die zweite Zeile des LCD
			lcd_puts("Freq: ");	
			for(i = 0; zeile2[i] && i < 10; i++)// 	auslesen bis NULL, 8 Zeichen nicht mehr
  			lcd_putc(zeile2[i]);
		
			silabs_setNewFreq (frequenz+10700000);	// nun die ausgewhlte Frequenz zum SI570
													// hier wird die ZF von 10.7 MHz addiert
			status.kanal_update = 0;
		}
	}
}

	// Belegung der Ports
	// Port B Bit 0 PB0 (PCINT0/CKL0/ICP1)											
	//        Bit 1 PB1 (OC1A/PCINT1)			Lebenslicht
	//        Bit 2 PB2	(SS/OC1B/PCINT3)		Ausgang PWM Signal fr S-Meter (WIP)
  	//        Bit 3 PB3	(MOSI/OC2A/PCINT3))		ISP MOSI
  	//        Bit 4 PB4	(MISO/PCINT4))			ISP MISO
  	//        Bit 5 PB5 (SCK/PCINT5)			ISP SCK
  	//        Bit 6 PB6	(PCINT6/XTAL1/TOSC1)	20MHz X-TAL
	//		  Bit 7 PB7	(PCINT7/XTAL2/TOSC2)	20MHz X-TAL

 	// Port C Bit 0 PC0 (ADC0/PCINT8)			Eingang RSSI Signal vom MC3371, fr S-Meter (WIP)
  	//        Bit 1 PC1	(ADC1/PCINT9)			-
 	//        Bit 2 PC2	(ADC2/PCINT10)			Eingang Druckknopf Drehimpulsgeber (WIP)
  	//        Bit 3 PC3	(ADC3/PCINT11)			Drehimpulsgeber A
  	//        Bit 4 PC4	(ADC4/SDA/PCINT12)		I2C SDA 
  	//        Bit 5 PC5	(ADC5/SCL/PCINT13)		I2C CLK 
  	//        Bit 6 PC6	(PCINT/RESET)			Reset
  			          										
	// Port D Bit 0 PD0	(PCINT16/RXD)			LCD E			
	//        Bit 1 PD1	(PCINT17/TXD)			LCD R/W
  	//        Bit 2 PD2	(PCINT18/INT0)			LCD RS
  	//        Bit 3 PD3	(PCINT19/OC2B/INT1)		Eingang Drehimpulsgeber B
  	//        Bit 4 PD4	(PCINT20/XCK/T0)		LCD Data dB4 	
  	//        Bit 5 PD5	(PCINT21/OC0B/T1)		LCD Data dB5 	
  	//        Bit 6 PD6	(PCINT22/OC0A/AIN0)		LCD Data dB6 	
  	//        Bit 7 PD7	(PCINT23/AIN1)			LCD Data dB7 			



